/*
Copyright © 2020 Roesner P/L

i4M Machine Controller
*/

var i2c = require('i2c-bus');
var fs = require('fs');


//log to console if it exists, else set console.log to nop
if (console){
	console.log('Console log enabled');
}
else {
	console.log = function(){};
}

//i2c comms to main board
var i2c1 = i2c.openSync(1);

//buffer for command data to main board
var txData1 = Buffer.alloc(32, 0);

var hardwareVer;
var softwareVer;
var cycleDataTimer;
var saveConfigTimer;
var saveDataTimer;
var networkName = '?';
var gpsHwVersion = '?';
var gpsSwVersion = '?';
var gpsPriorityNetwork = '?';
var gpsFallbackNetwork = '?';

var loadCellMcuOnline = true;
var mainMcuOnline = true;
var i2cLoadCellCounts = 0;
var i2cMainMcuCounts = 0;
var ledMode = 0;

var i2cLoadCellFlags = 0;
var i2cLoadCell0 = 0;
var i2cLoadCell1 = 0;
var i2cLoadCell2 = 0;
var i2cLoadCell3 = 0;
var resolvedLoadCell = 0;
var displayWeight = 0;

var i2cMainMcuFlags = 0;
var i2cSupplyVoltage = 0;
var i2cMeterSpeed = 0;
var i2cFanSpeed = 0;
var i2cGroundSpeed = 0;
var i2cValve1Command = 0;
var i2cTargetMeterSpeed = 0;

var masterOn = false;
var i2cMasterFlipFlop = 0;
var binEmpty = false;
var targetRate = 0;
var actualRate = 0;
var lastTargetRate = 0;
var rawRate = 0;
var badRateCounter = 0;
var fanRpm = 0;
var simGroundSpeedPps = 0;
var simGroundSpeedKph = 0;
var manualValve1 = 0;
var instantHectares = 0;
var overlapping = false;

var meterSpeedRatio = 0;
var pulsesPerHa = 0;

var gpsMsgCounter = 0;
var gpsComsTimer = 0;
var gpsLat = 0;
var gpsLon = 0;
var gpsLastLat = 0;
var gpsLastLon = 0;
var gpsSpeedKph = 0;
var gpsHeading = 0;
var gpsSpdArray = new Array(20).fill(0);

var alert1 = false;
var alert2 = false;
var alert3 = false;
var alert4 = false;
var alert5 = false;
var alert6 = false;
var alert7 = false;
var alert1Counter = 0;
var alert2Counter = 0;
var alert3Counter = 0;
var alert4Counter = 0;
var alert5Counter = 0;
var alert6Counter = 0;
var alert7Counter = 0;

//tracking filter
var deadBand = 0;
var featherTime = 0;
var fineAdjustStart = 0;
var fineAdjustUp = 0;
var fineAdjustDown = 0;
var coarseAdjust = 0;
var tooLowCounter = 0;
var tooHighCounter = 0;

//moving average filter
var sampleBuffer = new Array(100).fill(0);
var numberOfSamples = 100;
var samples = 0;
var samplePointer = 0;
var sampleTotal = 0;
var resetFilter = true;

//data file variables
var countdownKg = 0;
var calSampleKg = 0;
var hectareCounter = 0;

//config file variables
var loadCellsEnabled;
var quickFillKgs;
var useLoadCell0;
var useLoadCell1;
var useLoadCell2;
var useLoadCell3;
var loadCellZero;
var loadCellSpan;
var weightSpan;
var weightOffset;
var filterMode;
var valve1JumpStart;
var valve1Speed;
var valve1Brake;
var valve1TargetSize;
var distancePerPulse;
var gpsInstalled;
var gpsDistance;
var hectareFactor;
var storedProducts;
var storedCals;
var productPointer;
var productName;
var productCal;
var swathWidth;
var binSensorInstalled;
var fanInstalled;
var fanPulsesPerRev;
var fanLowAlert;
var fanHighAlert;


readWifiFile();
readVersionFile();
readDataStorageFile();
loadUserConfig();
//loadcell config not required for now, adc reads all 4 channels
//setTimeout(configLoadCellMcu, 1000);
//set controller led to solid green
ledMode = 2;
setTimeout(setLedMode, 1200);
setTimeout(configMainMcu, 1500);
setTimeout(cycleData, 2000);


function readWifiFile(){
	try {
		//read and parse the wifi network name file
		var rawData = fs.readFileSync('/home/pi/servermain/configWifi.json');
		var data = JSON.parse(rawData);
		
		networkName = data.networkName;
	}
	catch (e){
		console.log('Catch error readWifiFile');
	}
}

function readVersionFile(){
	try {
		//read and parse the version file
		var rawData = fs.readFileSync('/home/pi/servermain/version.json');
		var version = JSON.parse(rawData);
		
		hardwareVer = version.hardwareVer;
		softwareVer = version.softwareVer;
	}
	catch (e){
		console.log('Catch error readVersionFile');
	}
}

function readDataStorageFile(){
	try {
		//read and parse the dataStorage file
		var rawData = fs.readFileSync('/home/pi/servermain/appVR/dataStorage.json');
		var dataStorage = JSON.parse(rawData);
		
		countdownKg = dataStorage.countdownKg;
		calSampleKg = dataStorage.calSampleKg;
		hectareCounter = dataStorage.hectareCounter;
	}
	catch (e){
		console.log('Catch error readDataStorageFile');
	}
}

function saveDataStorageFile(){
	try {
		//load variables to save to dataStorage
		var dataStorage = {
			countdownKg: countdownKg,
			calSampleKg: calSampleKg,
			hectareCounter: hectareCounter
		};
		
		//convert to json strings and write to file
		var data = JSON.stringify(dataStorage, null, 2);
		fs.writeFileSync('/home/pi/servermain/appVR/dataStorage.json', data);
		
		console.log('Saving dataStorage file');
	}
	catch (e){
		console.log('Catch error saveDataStorageFile');
	}
}

function clearDataStorage(){
	try {
		//clear stored data
		countdownKg = 0;
		calSampleKg = 0;
		hectareCounter = 0;
		
		saveDataStorageFile();
	}
	catch (e){
		console.log('Catch error clearDataStorage');
	}
}

function loadUserConfig(){
	try {
		//read and parse the config file
		var rawData = fs.readFileSync('/home/pi/servermain/appVR/configUser.json');
		var config = JSON.parse(rawData);
		
		loadCellsEnabled = config.loadCellsEnabled;
		quickFillKgs = config.quickFillKgs;
		useLoadCell0 = config.useLoadCell0;
		useLoadCell1 = config.useLoadCell1;
		useLoadCell2 = config.useLoadCell2;
		useLoadCell3 = config.useLoadCell3;
		loadCellZero = config.loadCellZero;
		loadCellSpan = config.loadCellSpan;
		weightSpan = config.weightSpan;
		weightOffset = config.weightOffset;
		filterMode = config.filterMode;
		valve1JumpStart = config.valve1JumpStart;
		valve1Speed = config.valve1Speed;
		valve1Brake = config.valve1Brake;
		valve1TargetSize = config.valve1TargetSize;
		distancePerPulse = config.distancePerPulse;
		gpsInstalled = config.gpsInstalled;
		gpsDistance = config.gpsDistance;
		hectareFactor = config.hectareFactor;
		swathWidth = config.swathWidth;
		binSensorInstalled = config.binSensorInstalled;
		fanInstalled = config.fanInstalled;
		fanPulsesPerRev = config.fanPulsesPerRev;
		fanLowAlert = config.fanLowAlert;
		fanHighAlert = config.fanHighAlert;
		storedProducts = config.storedProducts;
		storedCals = config.storedCals;
		productPointer = config.productPointer;
		
		//set product variables based on the stored pointer
		productName = storedProducts[productPointer];
		productCal = storedCals[productPointer];

		console.log('Loading user config');
		console.log(config);
	}
	catch (e){
		//config file is blank or faulty, so load the factory data
		loadFactoryConfig();
		console.log('Catch error loadUserConfig');
	}
}

function loadFactoryConfig(){
	try {
		//read and parse the config file
		var rawData = fs.readFileSync('/home/pi/servermain/appVR/configFactory.json');
		var config = JSON.parse(rawData);
		
		loadCellsEnabled = config.loadCellsEnabled;
		quickFillKgs = config.quickFillKgs;
		useLoadCell0 = config.useLoadCell0;
		useLoadCell1 = config.useLoadCell1;
		useLoadCell2 = config.useLoadCell2;
		useLoadCell3 = config.useLoadCell3;
		loadCellZero = config.loadCellZero;
		loadCellSpan = config.loadCellSpan;
		weightSpan = config.weightSpan;
		weightOffset = config.weightOffset;
		filterMode = config.filterMode;
		valve1JumpStart = config.valve1JumpStart;
		valve1Speed = config.valve1Speed;
		valve1Brake = config.valve1Brake;
		valve1TargetSize = config.valve1TargetSize;
		distancePerPulse = config.distancePerPulse;
		gpsInstalled = config.gpsInstalled;
		gpsDistance = config.gpsDistance;
		hectareFactor = config.hectareFactor;
		swathWidth = config.swathWidth;
		binSensorInstalled = config.binSensorInstalled;
		fanInstalled = config.fanInstalled;
		fanPulsesPerRev = config.fanPulsesPerRev;
		fanLowAlert = config.fanLowAlert;
		fanHighAlert = config.fanHighAlert;
		storedProducts = config.storedProducts;
		storedCals = config.storedCals;
		productPointer = config.productPointer;
		
		//set product variables based on the stored pointer
		productName = storedProducts[productPointer];
		productCal = storedCals[productPointer];

		console.log('Loading factory config');
		console.log(config);
	}
	catch (e){
		console.log('Catch error loadFactoryConfig');
	}
}

function factoryLoadCells(){
	try {
		//read and parse the config file
		var rawData = fs.readFileSync('/home/pi/servermain/appMON/configFactory.json');
		var config = JSON.parse(rawData);
		
		useLoadCell0 = config.useLoadCell0;
		useLoadCell1 = config.useLoadCell1;
		useLoadCell2 = config.useLoadCell2;
		useLoadCell3 = config.useLoadCell3;
		loadCellZero = config.loadCellZero;
		loadCellSpan = config.loadCellSpan;
		weightSpan = config.weightSpan;
		weightOffset = config.weightOffset;
		filterMode = config.filterMode;

		console.log('Loading factory loadcells');
		console.log(config);
	}
	catch (e){
		console.log('Catch error loadFactoryLoadCell');
	}
}

function saveUserConfig(){
	try {
		//load variables used by pi and main board
		var configData = {
			loadCellsEnabled: loadCellsEnabled,
			quickFillKgs: quickFillKgs,
			useLoadCell0: useLoadCell0,
			useLoadCell1: useLoadCell1,
			useLoadCell2: useLoadCell2,
			useLoadCell3: useLoadCell3,
			loadCellZero: loadCellZero,
			loadCellSpan: loadCellSpan,
			weightSpan: weightSpan,
			weightOffset: weightOffset,
			filterMode: filterMode,
			valve1JumpStart: valve1JumpStart,
			valve1Speed: valve1Speed,
			valve1Brake: valve1Brake,
			valve1TargetSize: valve1TargetSize,
			distancePerPulse: distancePerPulse,
			gpsInstalled: gpsInstalled,
			gpsDistance: gpsDistance,
			hectareFactor: hectareFactor,
			swathWidth: swathWidth,
			binSensorInstalled: binSensorInstalled,
			fanInstalled: fanInstalled,
			fanPulsesPerRev: fanPulsesPerRev,
			fanLowAlert: fanLowAlert,
			fanHighAlert: fanHighAlert,
			storedProducts: storedProducts,
			storedCals: storedCals,
			productPointer: productPointer
		};
		
		//convert to json strings and write to file
		var data = JSON.stringify(configData, null, 2);
		fs.writeFileSync('/home/pi/servermain/appVR/configUser.json', data);
		
		console.log('Saving user config file');
	}
	catch (e){
		console.log('Catch error saveUserConfig');
	}
}

function saveFactoryConfig(){
	try {
		//load variables used by pi and main board
		var configData = {
			loadCellsEnabled: loadCellsEnabled,
			quickFillKgs: quickFillKgs,
			useLoadCell0: useLoadCell0,
			useLoadCell1: useLoadCell1,
			useLoadCell2: useLoadCell2,
			useLoadCell3: useLoadCell3,
			loadCellZero: loadCellZero,
			loadCellSpan: loadCellSpan,
			weightSpan: weightSpan,
			weightOffset: weightOffset,
			filterMode: filterMode,
			valve1JumpStart: valve1JumpStart,
			valve1Speed: valve1Speed,
			valve1Brake: valve1Brake,
			valve1TargetSize: valve1TargetSize,
			distancePerPulse: distancePerPulse,
			gpsInstalled: gpsInstalled,
			gpsDistance: gpsDistance,
			hectareFactor: hectareFactor,
			swathWidth: swathWidth,
			binSensorInstalled: binSensorInstalled,
			fanInstalled: fanInstalled,
			fanPulsesPerRev: fanPulsesPerRev,
			fanLowAlert: fanLowAlert,
			fanHighAlert: fanHighAlert,
			storedProducts: storedProducts,
			storedCals: storedCals,
			productPointer: productPointer
		};
		
		//convert to json strings and write to file
		var data = JSON.stringify(configData, null, 2);
		fs.writeFileSync('/home/pi/servermain/appVR/configFactory.json', data);
		
		console.log('Saving factory config file');
	}
	catch (e){
		console.log('Catch error saveFactoryConfig');
	}
}

function setLedMode(){
	try {
		//led modes 1 = solid red, 2 = solid green, 3 = flashing red, 4 = flashing green
		//command bytes to send via i2c
		txData1[0] = 0xc1;
		txData1[1] = 0x00;
		txData1[2] = 0x03;
		txData1[3] = ledMode;
		txData1[4] = 0x00;
		txData1[5] = 0x00;
		txData1[6] = 0x00;
		txData1[7] = 0x00;
		txData1[8] = 0x00;
		txData1[9] = 0x00;
		txData1[10] = 0x00;
		txData1[11] = 0x00;
		txData1[12] = 0x00;
		txData1[13] = 0x00;
		txData1[14] = 0x00;
		
		//send command to mcu
		writeMainMcu(txData1);
	}
	catch (e){
		console.log('Catch error setLedMode');
	}
}

function configLoadCellMcu(){
	try {
		var configFlags = 0;
		if (useLoadCell0 == true){
			configFlags = configFlags | 0x01;
		}
		if (useLoadCell1 == true){
			configFlags = configFlags | 0x02;
		}
		if (useLoadCell2 == true){
			configFlags = configFlags | 0x04;
		}
		if (useLoadCell3 == true){
			configFlags = configFlags | 0x08;
		}
		
		//command bytes to send via i2c
		txData1[0] = 0xc2;
		txData1[1] = 0x00;
		txData1[2] = 0x02;
		txData1[3] = 0x00;
		txData1[4] = byte0(configFlags);
		txData1[5] = 0x00;
		txData1[6] = 0x00;
		txData1[7] = 0x00;
		txData1[8] = 0x00;
		txData1[9] = 0x00;
		txData1[10] = 0x00;
		txData1[11] = 0x00;
		txData1[12] = 0x00;
		txData1[13] = 0x00;
		txData1[14] = 0x00;
		
		//send command to loadcell mcu
		writeLoadCellMcu(txData1);
	}
	catch (e){
		console.log('Catch error configLoadCellMcu');
	}
}


function configMainMcu(){
	try {
		//not required
	}
	catch (e){
		console.log('Catch error configMainMcu');
	}
}

var tStart = 0;
var tLast = 0;
var tDelta = 0;
var cyclesPerSec = 5;

//called regularly to get data from the main board
function cycleData(){
	try {
		//run again after this delay
		cycleDataTimer = setTimeout(cycleData, 200);
		
		//tStart = new Date().getTime();
		//tDelta = tStart - tLast;
		//console.log('cycle interval ' + tDelta);
		//tLast = tStart;
		
		/*
		//if the slave is online, read it
		if (loadCellMcuOnline){
			readLoadCellMcu();
		}
		
		//process loadcells if they are enabled
		//allow time for mainboard to respond min 5ms
		if (loadCellsEnabled){
			setTimeout(processLoadCells, 30);
		}
		*/

		//if the slave is online, read it
		if (mainMcuOnline){
			readMainMcu();
			//allow time for mainboard to respond min 5ms
			setTimeout(processMainMcu, 30);
		}
	}
	catch (e){
		console.log('Catch error cycleData');
	}
}

function readLoadCellMcu(){
	try {
		//message settings
		var slaveID = 80;
		var slaveRamPtr = Buffer.alloc(1,100);
		var packetLength = 15;
		var packet = Buffer.alloc(32, 0);

		//readI2cBlockSync no longer works since pi firmware and kernel updates
		//so use plain i2c commands to write and read seperately

		//write to slave to set ram pointer to the entry we want start reading from
		//then read a block of bytes
		i2c1.i2cWriteSync(slaveID,1,slaveRamPtr);
		i2c1.i2cReadSync(slaveID, packetLength, packet);
		
		//if valid end of frame byte
		//clr failure counter, and transfer the data
		if (packet[14] == 0xC3){
			loadCellMcuOnline = true;
			i2cLoadCellCounts ++;
			
			//console.log('I2C mcu2 read loadcells:   OK   ' + i2cLoadCellCounts);
			
			i2cLoadCellFlags = packet[1];
			i2cLoadCell0 = join3Bytes(packet[2], packet[3], packet[4]);
			i2cLoadCell1 = join3Bytes(packet[5], packet[6], packet[7]);
			i2cLoadCell2 = join3Bytes(packet[8], packet[9], packet[10]);
			i2cLoadCell3 = join3Bytes(packet[11], packet[12], packet[13]);
		}
		else {
			console.log('End of frame error readLoadCellMcu');
			//reset this flag after 2secs, this resets the mcu i2c module
			loadCellMcuOnline = false;
			setTimeout(function(){ loadCellMcuOnline = true; }, 2000);
		}
	}
	catch (e){
		console.log('Catch error readLoadCellMcu');
		//reset this flag after 2secs, this resets the mcu i2c module
		loadCellMcuOnline = false;
		setTimeout(function(){ loadCellMcuOnline = true; }, 2000);
	}
}

function processLoadCells(){
	try	{
		//sum the active loadcell channels and div by number of channels
		var average = 0;
		var channels = 0;
		var tempDisplayWeight = displayWeight;
		
		if (useLoadCell0 == true){
			average += i2cLoadCell0;
			channels ++;
		}
		if (useLoadCell1 == true){
			average += i2cLoadCell1;
			channels ++;
		}
		if (useLoadCell2 == true){
			average += i2cLoadCell2;
			channels ++;
		}
		if (useLoadCell3 == true){
			average += i2cLoadCell3;
			channels ++;
		}
		
		average = Math.round(average / channels);
		resolvedLoadCell = average;
		
		//if loadcell range is not zero, calc the weight
		if (loadCellSpan - loadCellZero){
			
			var newWeight = weightSpan * (resolvedLoadCell - loadCellZero);
			newWeight = newWeight / (loadCellSpan - loadCellZero);
			newWeight = parseInt(newWeight, 10);
			
			//apply the weight offset now
			newWeight += weightOffset;
			
			if (resetFilter){
				//reset data for moving ave filter
				sampleBuffer.fill(0);
				samples = 0;
				samplePointer = 0;
				sampleTotal = 0;
				resetFilter = false;
			}
			
			switch (filterMode){
				
				//filter off
				case 0:
					tempDisplayWeight = newWeight;
				break;
				
				
				//tracking filter, if display is too low, it will gradually increase
				//until feather time elapses, then apply coarse adjust
				case 1:
					featherTime = 15;
					deadBand = 5;
					fineAdjustStart = 4;
					fineAdjustUp += 2;
					fineAdjustDown += 2;
					coarseAdjust = 200;
					
					//too high
					if (displayWeight > (newWeight + deadBand))
					{
						tooHighCounter ++;
						tooLowCounter = 0;
						fineAdjustUp = 0;
						
						if (tooHighCounter < featherTime)
						{	
							tempDisplayWeight -= fineAdjustStart + fineAdjustDown;
						}
						else
						{
							tempDisplayWeight -= coarseAdjust;
							fineAdjustDown = 0;
							
							if (tempDisplayWeight < newWeight)
							{
								tempDisplayWeight = newWeight;
								tooHighCounter = 0;
							}
						}
					}
					//too low
					else if (displayWeight < (newWeight - deadBand))
					{
						tooLowCounter ++;
						tooHighCounter = 0;
						fineAdjustDown = 0;
						
						if (tooLowCounter < featherTime)
						{
							tempDisplayWeight += fineAdjustStart + fineAdjustUp;
						}
						else
						{
							tempDisplayWeight += coarseAdjust;
							fineAdjustUp = 0;
							
							if (tempDisplayWeight > newWeight)
							{
								tempDisplayWeight = newWeight;
								tooLowCounter = 0;
							}
						}
					}
					else
					{
						tooHighCounter = 0;
						tooLowCounter = 0;
						fineAdjustUp = 0;
						fineAdjustDown = 0;
					}
				break;
				
				
				//moving average filter
				case 2:
					numberOfSamples = 50;
					
					if (samples < numberOfSamples)
					{
						samples ++;
					}
					
					sampleTotal -= sampleBuffer[samplePointer];
					sampleTotal += newWeight;
					sampleBuffer[samplePointer] = newWeight;
					
					tempDisplayWeight = sampleTotal / samples;
					
					samplePointer ++;
					
					if (samplePointer == numberOfSamples)
					{
						samplePointer = 0;
					}
				break;
				
				default:
			}
			
			//if filter is on, snap to zero if within 50kg
			if (filterMode){
				if (tempDisplayWeight < 50 && tempDisplayWeight > -50){
					tempDisplayWeight = 0;
				}
			}
		}
		else {
			tempDisplayWeight = 0;
		}
		
		//if zero or span is the default value, load a big negative number
		//so the display prompts user to calibrate
		if (loadCellZero == 0 || loadCellSpan == 130000){
			displayWeight = -100000;
		}
		else {
			//update display, round to 10kg
			displayWeight = Math.round(tempDisplayWeight / 10) * 10;
		}
		
	}
	catch (e){
		console.log('Catch error processLoadCells');
	}
}

function writeLoadCellMcu(txData){
	try	{
		//message settings
		var slaveID = 80;
		var slaveRegister = 0;
		var packetLength = 15;
		
		//load end of frame byte
		txData[14] = 0xC3;
		
		//send block of bytes
		i2c1.writeI2cBlockSync(slaveID, slaveRegister, packetLength, txData);
		
		console.log('I2C loadcell mcu:          WRITE');
		
		//i2c1.closeSync();
	}
	catch (e){
		console.log('Catch error writeLoadCellMcu');
	}
}

function writeMainMcu(txData){
	try	{
		//message settings
		var slaveID = 60;
		var slaveRegister = 0;
		var packetLength = 15;
		
		//load end of frame byte
		txData[14] = 0xC3;
		
		//send block of bytes
		i2c1.writeI2cBlockSync(slaveID, slaveRegister, packetLength, txData);
		
		//console.log('I2C main mcu:              WRITE');
		
		//i2c1.closeSync();
	}
	catch (e){
		console.log('Catch error writeMainMcu');
	}
}

function readMainMcu(){
	try	{
		//message settings
		var slaveID = 60;
		var slaveRamPtr = Buffer.alloc(1,100);
		var packetLength = 30;
		var packet = Buffer.alloc(32, 0);

		//readI2cBlockSync no longer works since pi firmware and kernel updates
		//so use plain i2c commands to write and read seperately

		//write to slave to set ram pointer to the entry we want start reading from
		//then read a block of bytes
		i2c1.i2cWriteSync(slaveID,1,slaveRamPtr);
		i2c1.i2cReadSync(slaveID, packetLength, packet);
		
		//if valid end of frame byte
		//clr failure counter, and transfer the data
		if (packet[29] == 0xC3){
			mainMcuOnline = true;
			i2cMainMcuCounts ++;
				
			//console.log('I2C mcu1 read data1:       OK   ' + i2cMainMcuCounts);
			
			i2cMainMcuFlags = join2Bytes(packet[1], packet[2]);
			i2cSupplyVoltage = packet[3];

			//high speed sensor readings arrive as pulses per sec whole numbers
			//low speed arrive shifted left so we can divide to get decimal places
			//shifting right will give no decimals, so divide by the shift result instead
			i2cMeterSpeed = join2Bytes(packet[5], packet[6]) / (1 << packet[4]);
			i2cFanSpeed = join2Bytes(packet[8], packet[9]) / (1 << packet[7]);
			i2cGroundSpeed = join2Bytes(packet[11], packet[12]) / (1 << packet[10]);
			i2cTargetMeterSpeed = join2Bytes(packet[14], packet[15]) / (1 << packet[13]);
			i2cValve1Command = packet[16];

		}
		else {
			console.log('End of frame error readMainMcu');
			//reset this flag after 2secs, this resets the mcu i2c module
			mainMcuOnline = false;
			setTimeout(function(){ mainMcuOnline = true; }, 2000);
		}
	}
	catch (e){
		console.log('Catch error readMainMcu');
		//reset this flag after 2secs, this resets the mcu i2c module
		mainMcuOnline = false;
		setTimeout(function(){ mainMcuOnline = true; }, 2000);
	}
}


function processMainMcu(){
	try {
		/*
		//test values using wheel speed
		if (masterOn){
			i2cMeterSpeed = 150;
			i2cFanSpeed = 30;
			i2cGroundSpeed = 10;
			i2cTargetMeterSpeed = 715;
		}
		else {
			i2cMeterSpeed = 0;
			i2cFanSpeed = 30;
			i2cGroundSpeed = 0;
		}
		
		//test values using gps speed
		if (masterOn){
			i2cMeterSpeed = 150;
			i2cFanSpeed = 30;
			gpsSpeedKph = 15;
			i2cTargetMeterSpeed = 715;
		}
		else {
			i2cMeterSpeed = 0;
			i2cFanSpeed = 30;
			gpsSpeedKph = 15;
		}
		*/
		//main mcu sends a flag to identify when the master switch has been pressed
		//the flag just flip flops, it doesn't mean the master is on or off
		//the pi and tablet checks this flag, if it's changed, it knows it was pressed
		//get the initial value first, then after a few read cycles, allow master toggle
		var newI2cFlipFlop = 0;
		if (i2cMainMcuFlags & 0x10){
			newI2cFlipFlop = 1;
		}
		
		if (i2cMainMcuCounts > 5){
			
			if (i2cMasterFlipFlop != newI2cFlipFlop){
				//if it's changed, store the new value and toggle the master
				i2cMasterFlipFlop = newI2cFlipFlop;
				masterOn = !masterOn;
				
				//save the data storage file
				if (masterOn == false){
					saveDataStorageFile();
				}
				
				setMeterRatio();
			}
		}
		else {
			i2cMasterFlipFlop = newI2cFlipFlop;
		}

		if (i2cMainMcuFlags & 0x04){
			binEmpty = true;
		} else {
			binEmpty = false;
		}
		
		//calc actual rate to display
		calcActualRate();
		
		//calc hectares to add to tablet hectare counter
		calcHectareCounter();
		
		//fan speed rpm rounded to nearest 10, i2cFanSpeed is pulses per sec
		fanRpm = Math.round((i2cFanSpeed * 6) / fanPulsesPerRev) * 10;
		
		//check if we need to fire any alerts
		processAlerts();
		
		
		if (gpsInstalled){
			
			//gps speed arrives as km/h, we need to send pulsesPerSec x128
			//use a fixed distance per pulse to generate 10 pulses per metre
			//so gpsSpeedPps = (gpsSpeedKph x 1000metres x 10pulsesPerMetre x 128) / 3600secondsPerHour
			var gpsSpeedPps = (gpsSpeedKph * 1280) / 3.6;
			
			//command bytes to send via i2c
			txData1[0] = 0xc2;
			txData1[1] = 0x00;
			txData1[2] = 0x11;
			txData1[3] = byte1(gpsSpeedPps);
			txData1[4] = byte0(gpsSpeedPps);
			txData1[5] = 0x01;
			txData1[6] = 0x00;
			txData1[7] = 0x00;
			txData1[8] = 0x00;
			txData1[9] = 0x00;
			txData1[10] = 0x00;
			txData1[11] = 0x00;
			txData1[12] = 0x00;
			txData1[13] = 0x00;
			txData1[14] = 0x00;
			
			//send command to mcu
			writeMainMcu(txData1);
		}
		else {
			//command bytes to send via i2c
			txData1[0] = 0xc2;
			txData1[1] = 0x00;
			txData1[2] = 0x11;
			txData1[3] = 0x00;
			txData1[4] = 0x00;
			txData1[5] = 0x00;
			txData1[6] = 0x00;
			txData1[7] = 0x00;
			txData1[8] = 0x00;
			txData1[9] = 0x00;
			txData1[10] = 0x00;
			txData1[11] = 0x00;
			txData1[12] = 0x00;
			txData1[13] = 0x00;
			txData1[14] = 0x00;
			
			//send command to mcu
			writeMainMcu(txData1);
		}
		
		
		
		
		//if something loaded the data storage counter, decrement it, when it hits zero, save the file
		if (saveDataTimer > 0){
			
			//console.log('save data timer: ' + saveDataTimer);
			
			saveDataTimer --;
			if (saveDataTimer == 0){
				saveDataStorageFile();
			}
		}
	}
	catch (e){
		console.log('Catch error processMainMcu');
	}
}


function calcActualRate(){
	try	{
		//calc the kg dispensed per second, i2cMeterSpeed is pulses per sec
		var kgDispensed = i2cMeterSpeed / (productCal * cyclesPerSec);
		
		//if we have dispensed kgs, update the cal sample weight, load counter to save the data
		if (kgDispensed > 0){
			calSampleKg += kgDispensed;
			saveDataTimer = cyclesPerSec * 3;
		}
		
		//if not using loadcells, process the kg counter
		if (!loadCellsEnabled){
			
			countdownKg -= kgDispensed;
			
			if (countdownKg < 0){
				countdownKg = 0;
				displayWeight = 0;
			}
			else {
				//update display, round to 10kg
				displayWeight = Math.round(countdownKg * 0.1) * 10;
			}
		}
		
		//calc the rate based on hectares covered, using sim or gps or wheel groundspeed
		if (simGroundSpeedPps > 0){
			var haCovered = (simGroundSpeedPps * distancePerPulse * swathWidth * hectareFactor) * 0.0001 / cyclesPerSec;
			if (haCovered){
				rawRate = Math.round(kgDispensed / haCovered);
			}
		}
		//gps speed is km/h
		else if (gpsSpeedKph > 0){
			var haCovered = (gpsSpeedKph * swathWidth * hectareFactor) / (36000 * cyclesPerSec);
			if (haCovered){
				rawRate = Math.round(kgDispensed / haCovered);
			}
		}
		else if (i2cGroundSpeed > 0){
			var haCovered = (i2cGroundSpeed * distancePerPulse * swathWidth * hectareFactor) * 0.0001 / cyclesPerSec;
			if (haCovered){
				rawRate = Math.round(kgDispensed / haCovered);
			}
		}
		else {
			rawRate = 0;
		}
		
		//when the raw rate is close to the target rate, the display will snap to the target rate
		//if target rate is zero or raw rate is zero, don't snap to rate
		if (targetRate == 0 || rawRate == 0){
			actualRate = rawRate;
			badRateCounter = 100;
		}
		//if target rate has changed, don't snap to rate
		else if (targetRate != lastTargetRate){
			actualRate = rawRate;
			badRateCounter = 100;
		}
		else {
			//if within 5% of target clr the counter, else inc the counter
			var rawRatio = rawRate / targetRate;
			if (rawRatio < 1.05 && rawRatio > 0.95){
				badRateCounter = 0;
			}
			else {
				badRateCounter ++;
			}
			
			//if beyond threshold for 6sec, show raw rate, else snap to target rate
			if (badRateCounter > 30){
				actualRate = rawRate;
			}
			else {
				actualRate = targetRate;
			}
		}
		
		//keep track of the current target rate
		lastTargetRate = targetRate;
		
	}
	catch (e){
		console.log('Catch error calcActualRate');
	}
}


function calcHectareCounter(){
	try	{
		//calc hectares to add to tablet hectare counter
		instantHectares = 0;

		if (gpsInstalled){
			if (masterOn && gpsSpeedKph && !overlapping){
				instantHectares = (gpsSpeedKph * swathWidth * hectareFactor) / (36000 * cyclesPerSec);
			}
		}
		else if (masterOn && i2cGroundSpeed){
			instantHectares = (i2cGroundSpeed * distancePerPulse * swathWidth * hectareFactor) / (10000 * cyclesPerSec);
		}

		//accumulate hectare counter
		hectareCounter += instantHectares;
		
		//if we have instant hectares, the data has changed, so load a counter to save the data storage file
		if (instantHectares > 0){
			saveDataTimer = cyclesPerSec * 3;
		}
	}
	catch (e){
		console.log('Catch error calcHectareCounter');
	}
}


function setMeterRatio(){
	try {
		//the mainboard processor adjusts the meter pulses per second based on 
		//groundspeed pulses per second, so we just send it a ratio value
		//it makes adjustments at 10Hz
		
		//if master is on, calc the meter speed ratio, else set it to zero
		if (masterOn == true && targetRate != 0){
			
			var pulsesPerTargetMass = productCal * targetRate;
			
			//for gps speed we simulate 10 pulses per metre, so distance per pulse is 0.1
			//else wheel speed uses the user setting distancePerPulse
			if (gpsInstalled){
				pulsesPerHa = 10000 / (swathWidth * 0.1 * hectareFactor);
				meterSpeedRatio = pulsesPerTargetMass / pulsesPerHa;
			}
			else {
				pulsesPerHa = 10000 / (swathWidth * distancePerPulse * hectareFactor);
				meterSpeedRatio = pulsesPerTargetMass / pulsesPerHa;
			}
		}
		else {
			meterSpeedRatio = 0;
		}

		//we need to encode the value before packing into bytes
		var meterSpdRatioShiftedLeft = 0;
		var encodedMeterSpeedRatio = meterSpeedRatio;

		//we'll send as 16bit to main mcu, so cap if greater than 16bit
		if (encodedMeterSpeedRatio > 65535){
			encodedMeterSpeedRatio = 65535;
		}

		//if less than 500 encode it as low speed mode
		//x128 so the main mcu can right shift to get decimals
		if (encodedMeterSpeedRatio < 500){
			encodedMeterSpeedRatio *= 128;
			meterSpdRatioShiftedLeft = 7;
		}



		//we need to encode the value before packing into bytes
		var simGroundSpdShiftedLeft = 0;
		var encodedSimGroundSpeedPps = simGroundSpeedPps;

		//we'll send as 16bit to main mcu, so cap if greater than 16bit
		if (encodedSimGroundSpeedPps > 65535){
			encodedSimGroundSpeedPps = 65535;
		}
		
		//if less than 500 encode it as low speed mode
		//x128 so the main mcu can right shift to get decimals
		if (encodedSimGroundSpeedPps < 500){
			encodedSimGroundSpeedPps *= 128;
			simGroundSpdShiftedLeft = 7;
		}


		//command bytes to send via i2c
		txData1[0] = 0xc2;
		txData1[1] = 0x00;
		txData1[2] = 0x12;
		txData1[3] = meterSpdRatioShiftedLeft;
		txData1[4] = byte1(encodedMeterSpeedRatio);
		txData1[5] = byte0(encodedMeterSpeedRatio);
		txData1[6] = valve1Speed;
		txData1[7] = valve1Brake;
		txData1[8] = valve1TargetSize;
		txData1[9] = (valve1JumpStart * 2.55) & 0xff;
		txData1[10] = (manualValve1 * 2.55) & 0xff;
		txData1[11] = simGroundSpdShiftedLeft;
		txData1[12] = byte1(encodedSimGroundSpeedPps);
		txData1[13] = byte0(encodedSimGroundSpeedPps);
		txData1[14] = 0x00;

		/*
		console.log(' ');
		console.log('meterSpeedRatio ' + meterSpeedRatio);
		console.log('meterSpdRatioShiftedLeft ' + meterSpdRatioShiftedLeft);
		console.log('encodedMeterSpeedRatio ' + encodedMeterSpeedRatio);
		console.log('simGroundSpeedPps ' + simGroundSpeedPps);
		console.log('simGroundSpdShiftedLeft ' + simGroundSpdShiftedLeft);
		console.log('encodedSimGroundSpeedPps ' + encodedSimGroundSpeedPps);
		console.log(' ');
		*/

		
		//send command to mcu
		writeMainMcu(txData1);
	}
	catch (e){
		console.log('Catch error setMeterRatio');
	}
}

//text for each alert is preset in tablet app
//repeat at 5times per second, so 5counts is 1sec
function processAlerts(){
	try	{
		//alert2 meter is turning, when master switch is off or target is zero
		if (!masterOn || targetRate == 0){
			
			if (i2cMeterSpeed){
				alert2Counter ++;
				
				//fire the alert
				if (alert2Counter > 30){
					alert2 = true;
				}
			}
			else {
				alert2Counter = 0;
				alert2 = false;
			}
		}
		else {
			alert2Counter = 0;
			alert2 = false;
		}
		
		//alert3 output rate not matching the target rate
		if (masterOn && targetRate && (gpsSpeedKph || i2cGroundSpeed)){
			var rateRatio = actualRate / targetRate;
			
			if (rateRatio < 0.8 || rateRatio > 1.2){
				alert3Counter ++;
			}
			else {
				alert3Counter = 0;
				alert3 = false;
			}
			
			//fire the alert
			if (alert3Counter > 40){
				alert3 = true;
			}
		}
		else {
			alert3Counter = 0;
			alert3 = false;
		}
		
		//alert4 fan speed too low
		if (masterOn && i2cMeterSpeed){
			if (fanRpm < fanLowAlert){
				alert4Counter ++;
			}
			else {
				alert4Counter = 0;
				alert4 = false;
			}
			
			//fire the alert
			if (alert4Counter > 10){
				alert4 = true;
			}
		}
		else {
			alert4Counter = 0;
			alert4 = false;
		}
		
		//alert5 fan speed too high
		if (masterOn && i2cMeterSpeed){
			if (fanRpm > fanHighAlert){
				alert5Counter ++;
			}
			else {
				alert5Counter = 0;
				alert5 = false;
			}
			
			//fire the alert
			if (alert5Counter > 10){
				alert5 = true;
			}
		}
		else {
			alert5Counter = 0;
			alert5 = false;
		}

		//alert6 bin is empty
		if (binSensorInstalled && binEmpty){
			alert6Counter ++;
			
			//fire the alert
			if (alert6Counter > 30){
				alert6 = true;
			}
		}
		else {
			alert6Counter = 0;
			alert6 = false;
		}
		
		
	}
	catch (e){
		console.log('Catch error processAlerts');
	}
}



//concatena bytes and return hex value
function join2Bytes(b1, b0){
	var x1 = padByte(b1) + padByte(b0);
	x1 = parseInt(x1, 16);
	return x1;
}

//concatena 3 bytes and return hex value
function join3Bytes(b2, b1, b0){
	var x1 = padByte(b2) + padByte(b1) + padByte(b0);
	x1 = parseInt(x1, 16);
	return x1;
}

//concatena 4 bytes and return hex value
function join4Bytes(b3, b2, b1, b0){
	var x1 = padByte(b3) + padByte(b2) + padByte(b1) + padByte(b0);
	x1 = parseInt(x1, 16);
	return x1;
}


//converts byte to string with padded zero
function padByte(b){
	if (b < 0x10){
		b = '0' + b.toString(16);
	}
	else {
		b = b.toString(16);
	}
	return b;
}

//extract lowest byte
function byte0(bb){
	var b0 = bb & 0xFF;
	return b0;
}

//discard lowest byte, extract next byte
function byte1(bb){
	var b1 = (bb >> 8) & 0xFF;
	return b1;
}

//discard lowest two bytes, extract next byte
function byte2(bb){
	var b2 = (bb >> 16) & 0xFF;
	return b2;
}





//__________________________________________________________________________________________
//server for webpages, sockets, and http requests

var express = require('express');  
var app = express();
var server = require('http').createServer(app);
var io = require('socket.io')(server);

//when express receives a get request, send the index file
app.use(express.static(__dirname)); 
app.get('/', function(request, response, next) 
{	
	console.log('Received http get request');
	
    response.sendFile(__dirname + '/index.html');
});


//when express receives a post request, extract the data and respond
app.use(express.json());
app.use(express.urlencoded({ extended: true })); 
app.post('/', function(request, response, next) 
{	
	var bodyData = request.body;
	
	if (bodyData.command == 'refreshData'){
		
		//console.log('Http post request: ' + JSON.stringify(bodyData));
		
		//parse values from the body data
		var tabletTargetRate = parseFloat(bodyData.targetRate);
		
		//coverage overlap
		if (bodyData.overlapping == '1'){
			overlapping = true;
		}
		else {
			overlapping = false;
		}
		
		//tablet master switch was pressed
		if (bodyData.masterPressed == '1'){
			//update master status, if turning on also set the door
			if (bodyData.implementOn == '1'){
				masterOn = true;
			}
			//else if turning off, set flag and save data storage file
			else {
				masterOn = false;
				saveDataStorageFile();
			}
			
			//update other values and send
			targetRate = tabletTargetRate;
			setMeterRatio();
		}
		//else if rate has changed update values
		else if (targetRate != tabletTargetRate){
			targetRate = tabletTargetRate;
			setMeterRatio();
		}
	}
	
	//send new data to to tablet
	response.send({
		implementOn: masterOn,
		actualRate: actualRate,
		displayWeight: displayWeight,
		fanInstalled: fanInstalled,
		fanRpm: fanRpm,
		hectareCounter: hectareCounter,
		swathWidth: swathWidth,
		gpsDistance: gpsDistance,
		gpsLat: gpsLat,
		gpsLon: gpsLon,
		alert1: alert1,
		alert2: alert2,
		alert3: alert3,
		alert4: alert4,
		alert5: alert5,
		alert6: alert6,
		alert7: alert7
	});
});


//socket on, notify when socket is connected
io.on('connection', function(socket){
	
	console.log('Socket is connected');
	
	//simple message from webpage
	socket.on('join', function(data){
		console.log(data);
		socket.emit('status', 'Online');	
	});
	
	//__________________________________________________________________________________________
	//gps receiver
	
	//incoming request from gps to check if another gps unit is already connected to this controller
	//if gpsMsgCounter is zero, there are no other receivers connected, so it's ok to connect
	socket.on('gpsConnectedReq', function(data){
		socket.emit('gpsConnectedRes',{
			gpsMsgCounter: gpsMsgCounter
		});
	});
	
	//incoming gps data
	socket.on('gpsData', function(data){
		
		if (gpsInstalled){
			//we received gps data so kill the message timeout check
			clearTimeout(gpsComsTimer);
			gpsMsgCounter ++;
			
			//get version and heading
			gpsHeading = data.gpsHeading;
			gpsHwVersion = data.hardwareVer;
			gpsSwVersion = data.softwareVer;
			gpsPriorityNetwork = data.priorityNetwork;
			gpsFallbackNetwork = data.fallbackNetwork;
			
			//ignore small lat lon shifts, keep track of the last position we sent to tablet and only update it
			//when the newest lat lon has moved by more than the min shift amount, 0.00001 = 1.1m
			var minShift = 0.000005;
			var deltaLat = data.gpsLat - gpsLastLat;
			var deltaLon = data.gpsLon - gpsLastLon;
			
			if (deltaLat > minShift || deltaLat < -minShift || deltaLon > minShift || deltaLon < -minShift){
				gpsLat = data.gpsLat;
				gpsLon = data.gpsLon;
				gpsLastLat = gpsLat;
				gpsLastLon = gpsLon;
			}
			
			//gps speed is erratic, can jump by 10% so use a 4sec 20sample moving ave filter
			var pointer = gpsSpdArray.length - 1;
			var sum = 0;
			//move all entries up one, discarding the last entry, keep track of the array sum
			while (pointer > 0){
				gpsSpdArray[pointer] = gpsSpdArray[pointer - 1];
				sum += gpsSpdArray[pointer];
				pointer --;
			}
			
			//load latest gps speed to first entry and update the sum
			gpsSpdArray[0] = data.gpsSpeedKph;
			sum += gpsSpdArray[0];
			//get average of the 20 samples
			var newGpsSpd = sum * 0.05;
			
			//ignore small speed values, prevents meter movement while standing still
			//also ignore large speed values to prevent ha counter jumps
			if (newGpsSpd > 0.9 && newGpsSpd < 80){
				gpsSpeedKph = newGpsSpd;
			} else {
				gpsSpeedKph = 0;
			}
			
			//if we don't receive new gps updates, set gps data to zero
			gpsComsTimer = setTimeout(function(){
				gpsLat = 0;
				gpsLon = 0;
				gpsSpeedKph = 0;
				gpsHeading = 0;
				gpsMsgCounter = 0;
				console.log('no gps message');
			}, 3000);
		}
	});
	
	//__________________________________________________________________________________________
	//index page
	
	//webpage requesting data from the pi
	socket.on('indexPageReq', function(data){
		socket.emit('indexPageRes',{
			loadCellsEnabled: loadCellsEnabled,
			displayWeight: displayWeight,
			hectareCounter: hectareCounter,
			productName: productName
		});
	});
	
	//__________________________________________________________________________________________
	//products
	
	//webpage requesting data from the pi
	socket.on('productsPageReq', function(data){
		socket.emit('productsPageRes',{
			productName: productName,
			productCal: productCal,
			calSampleKg: calSampleKg
		});
	});
	
	//inc the product pointer, reset the timeout to save the config file so
	//we store the active product
	socket.on('nextProduct', function (data){
		
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		productPointer ++;
		if (productPointer > 19){
			productPointer = 0;
		}
		
		productName = storedProducts[productPointer];
		productCal = storedCals[productPointer];
	});
	
	//dec the product pointer, reset the timeout to save the config file so
	//we store the active product
	socket.on('prevProduct', function (data){
		
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		productPointer --;
		if (productPointer < 0){
			productPointer = 19;
		}
		
		productName = storedProducts[productPointer];
		productCal = storedCals[productPointer];
	});
	
	//user updated product info, save it
	socket.on('productSaveName', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		productName = data;
		storedProducts[productPointer] = productName;
	});
	
	//user updated product info, save it
	socket.on('productSaveCal', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		productCal = data;
		storedCals[productPointer] = productCal;
	});
	
	//start new cal sample
	socket.on('newCalSample', function (data){
		//zero the sample, refresh cal factor and door height
		calSampleKg = 0;
		saveDataStorageFile();
		productCal = storedCals[productPointer];
		//set the meter to refresh pulses per kg etc
		setMeterRatio();
	});
	
	//user entered a cal sample weight
	socket.on('calSampleWeight', function (data){
		//if the entered kg and counted kg are valid
		if (data > 0 && calSampleKg > 0){
			clearTimeout(saveConfigTimer);
			saveConfigTimer = setTimeout(saveUserConfig, 4000);
			
			//the correction factor is the counted kg divided by the actual kg
			var calAdjust = calSampleKg / data;
			var adjustedCal = productCal * calAdjust;
			
			//use math round to set 3 decimal places, this method is fastest
			productCal = Math.round(adjustedCal * 1000) / 1000;
			storedCals[productPointer] = productCal;
		}
		//update to show the entered kg
		calSampleKg = data;
		saveDataStorageFile();
	});
	
	//__________________________________________________________________________________________
	//settings machine
	
	//webpage requesting data from the pi
	socket.on('setupMachineReq', function(data){
		//calc groundspeed to display on settings page
		var groundSpeedKph = i2cGroundSpeed * distancePerPulse * 3.6;
			
		socket.emit('setupMachineRes',{
			loadCellsEnabled: loadCellsEnabled,
			swathWidth: swathWidth,
			distancePerPulse: distancePerPulse,
			groundSpeedKph: groundSpeedKph,
			valve1JumpStart: valve1JumpStart,
			valve1Speed: valve1Speed,
			valve1Brake: valve1Brake,
			valve1TargetSize: valve1TargetSize,
			gpsInstalled: gpsInstalled,
			binSensorInstalled: binSensorInstalled,
			fanInstalled: fanInstalled,
			fanPulsesPerRev: fanPulsesPerRev,
			fanLowAlert: fanLowAlert,
			fanHighAlert: fanHighAlert
		});
	});
	
	//set machine width
	socket.on('setSwathWidth', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		swathWidth = data;
		//run set meter to update other output data
		setMeterRatio();
	});
	
	//wheel sensor cal
	socket.on('wheelCal', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		distancePerPulse = data;
		//run set meter to update other output data
		setMeterRatio();
	});
	
	//change bin low sensor installed
	socket.on('changeBinSensorInstalled', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		binSensorInstalled = !binSensorInstalled;
	});
	
	//change fan installed
	socket.on('changeFanInstalled', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		fanInstalled = !fanInstalled;
	});
	
	//set fan sensor pulses per rev
	socket.on('setFanPulses', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		fanPulsesPerRev = data;
	});

	//set fan rpm low alert
	socket.on('setFanLowAlert', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		fanLowAlert = data;
	});

	//set fan rpm high alert
	socket.on('setFanHighAlert', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		fanHighAlert = data;
	});

	//set valve jump start
	socket.on('valveJumpStart', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		valve1JumpStart = data;
		//run set belt to update other output data
		setMeterRatio();
	});
	
	//set valve speed
	socket.on('valveSpeed', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		valve1Speed = data;
		//run set belt to update other output data
		setMeterRatio();
	});
	
	//set valve damping
	socket.on('valveDamping', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		valve1Brake = data;
		//run set belt to update other output data
		setMeterRatio();
	});
	
	//set valve green zone
	socket.on('valveZone', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		valve1TargetSize = data;
		//run set belt to update other output data
		setMeterRatio();
	});
	
	
	//__________________________________________________________________________________________
	//settings device
	
	//webpage requesting data from the pi
	socket.on('setupDeviceReq', function(data){
		socket.emit('setupDeviceRes',{
			loadCellsEnabled: loadCellsEnabled,
			networkName: networkName
		});
	});
	
	//reset to factory settings
	socket.on('factoryReset', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		loadFactoryConfig();
		clearDataStorage();
	});
	
	
	//change network name
	socket.on('setNetworkName', function (data){
		
		//set the new network name
		networkName = data;
		
		//save this in it's own file so all controller apps have access to it
		var netNameJson = {
			networkName: networkName
		};
		//convert to json string and write to the servermain config file
		var netNameStr = JSON.stringify(netNameJson, null, 2);
		fs.writeFileSync('/home/pi/servermain/configWifi.json', netNameStr);
		
		//broadcast the network name has changed so networked devices can switch over
		socket.broadcast.emit('networkNameChanged',{
			networkName: networkName
		});
		
		//then rewrite the wlan0 setup file, see usb image notes for details
		try {
			
			var string = 'interface=wlan0\n' +
				'driver=nl80211\n' +
				'ssid=' + networkName + '\n' +
				'hw_mode=g\n' +
				'channel=6\n' +
				'macaddr_acl=0\n' +
				'auth_algs=1\n' +
				'ignore_broadcast_ssid=0';
			
			fs.writeFileSync('/etc/hostapd/hostapd.conf', string);
		}
		catch (e){
			console.log('Catch error setNetworkName' + e);
		}
		
		//reboot the pi
		setTimeout(function(){
			var execute = require('child_process').exec;
			execute('sudo /sbin/reboot');
		}, 4000);
	});
	
	
	//__________________________________________________________________________________________
	//settings factory
	
	//webpage requesting data from the pi
	socket.on('factoryPage', function(data){
		//nothing required
	});
	
	//save to factory settings
	socket.on('saveFactory', function (data){		
		saveFactoryConfig();
	});
	
	//__________________________________________________________________________________________
	//settings hopper without loadcells
	
	//webpage requesting data from the pi
	socket.on('setupKgCounterReq', function(data){
		socket.emit('setupKgCounterRes',{
			loadCellsEnabled: loadCellsEnabled,
			quickFillKgs: quickFillKgs
		});
	});
	
	//set hopper kgs
	socket.on('setHopperKg', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		countdownKg = data;
		quickFillKgs = data;
		displayWeight = Math.round(countdownKg / 10) * 10;
		saveDataStorageFile();
	});
	
	//quick fill hopper kgs to previous amount
	socket.on('quickFillHopperKg', function (data){
		countdownKg = quickFillKgs;
		displayWeight = Math.round(countdownKg / 10) * 10;
		saveDataStorageFile();
	});
	
	//enable loadcells
	socket.on('enableLoadCells', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		loadCellsEnabled = true;
		resetFilter = true;
	});
	
	//__________________________________________________________________________________________
	//settings hopper with loadcells
	
	//webpage requesting data from the pi
	socket.on('setupLoadCellsReq', function(data){
		socket.emit('setupLoadCellsRes',{
			loadCellsEnabled: loadCellsEnabled,
			useLoadCell0: useLoadCell0,
			useLoadCell1: useLoadCell1,
			useLoadCell2: useLoadCell2,
			useLoadCell3: useLoadCell3,
			loadCellZero: loadCellZero,
			loadCellSpan: loadCellSpan,
			weightSpan: weightSpan,
			weightOffset: weightOffset,
			filterMode: filterMode,
			displayWeight: displayWeight
		});
	});
	
	//quick zero, set a weight offset
	socket.on('cellQuickZero', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		weightOffset = weightOffset + (0 - displayWeight);
		displayWeight = 0;
	});
	
	//cal zero
	socket.on('cellCalZero', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		weightOffset = 0;
		displayWeight = 0;
		loadCellZero = resolvedLoadCell;
		resetFilter = true;
	});
	
	//cal span
	socket.on('cellCalSpan', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		weightSpan = data;
		loadCellSpan = resolvedLoadCell;
		displayWeight = data;
		resetFilter = true;
	});
	
	//set loadcells to factory cal
	socket.on('factoryLoadCells', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		factoryLoadCells();
	});
	
	//set filter mode
	socket.on('setFilter', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		filterMode ++;
		if (filterMode > 2)
		{
			filterMode = 0;
		}
		
		resetFilter = true;
	});
	
	//set channel on/off
	socket.on('cellChannel0', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		useLoadCell0 = !useLoadCell0;
	});
	
	//set channel on/off
	socket.on('cellChannel1', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		useLoadCell1 = !useLoadCell1;
	});
	
	//set channel on/off
	socket.on('cellChannel2', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		useLoadCell2 = !useLoadCell2;
	});
	
	//set channel on/off
	socket.on('cellChannel3', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		useLoadCell3 = !useLoadCell3;
	});
	
	//disable loadcells
	socket.on('disableLoadCells', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		loadCellsEnabled = false;
	});
	
	
	//__________________________________________________________________________________________
	//settings hectare counter
	
	//webpage requesting data from the pi
	socket.on('setupHectaresReq', function(data){
		socket.emit('setupHectaresRes',{
			loadCellsEnabled: loadCellsEnabled,
			hectareCounter: hectareCounter
		});
	});
	
	//clear hectare counter
	socket.on('zeroHectares', function (data){
		hectareCounter = 0;
		saveDataStorageFile();
	});
	
	//calibrate hectare counter
	socket.on('calHectares', function (data){
		var newHectares = data;
		if (hectareCounter > 0){
			
			hectareFactor = newHectares / hectareCounter;
			
			clearTimeout(saveConfigTimer);
			saveConfigTimer = setTimeout(saveUserConfig, 4000);
		}
		
		hectareCounter = newHectares;
		saveDataStorageFile();
	});
		
		
	//__________________________________________________________________________________________
	//settings gps
	
	//webpage requesting data from the pi
	socket.on('setupGpsReq', function(data){
		socket.emit('setupGpsRes',{
			loadCellsEnabled: loadCellsEnabled,
			gpsInstalled: gpsInstalled,
			gpsDistance: gpsDistance
		});
	});
	
	//toggle gps installed
	socket.on('changeGps', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		gpsInstalled = !gpsInstalled;
		
		//if not installed, reset gps data to zero
		if (!gpsInstalled){
			gpsLat = 0;
			gpsLon = 0;
			gpsSpeedKph = 0;
			gpsHeading = 0;
			gpsMsgCounter = 0;
		}
	});
	
	//set gps distance
	socket.on('setGpsDistance', function (data){
		clearTimeout(saveConfigTimer);
		saveConfigTimer = setTimeout(saveUserConfig, 4000);
		
		gpsDistance = data;
	});
	
	
	//__________________________________________________________________________________________
	//diag system general
	
	//webpage requesting data from the pi
	socket.on('diagSystemReq', function(data){
		socket.emit('diagSystemRes',{
			loadCellMcuOnline: loadCellMcuOnline,
			mainMcuOnline: mainMcuOnline,
			i2cLoadCellCounts: i2cLoadCellCounts,
			i2cMainMcuCounts: i2cMainMcuCounts,
			i2cSupplyVoltage: i2cSupplyVoltage,
			hardwareVer: hardwareVer,
			softwareVer: softwareVer
		});
	});
	
	//__________________________________________________________________________________________
	//diag gps
	
	//webpage requesting data from the pi
	socket.on('diagGpsReq', function(data){
		socket.emit('diagGpsRes',{
			gpsInstalled: gpsInstalled,
			gpsMsgCounter: gpsMsgCounter,
			gpsLat: gpsLat,
			gpsLon: gpsLon,
			gpsSpeedKph: gpsSpeedKph,
			gpsHeading: gpsHeading,
			gpsSwVersion: gpsSwVersion,
			gpsHwVersion: gpsHwVersion,
			gpsPriorityNetwork: gpsPriorityNetwork,
			gpsFallbackNetwork: gpsFallbackNetwork
		});
	});
	
	//__________________________________________________________________________________________
	//diag loadcells
	
	//webpage requesting data from the pi
	socket.on('diagLoadCellsReq', function(data){
		socket.emit('diagLoadCellsRes',{
			i2cLoadCellFlags: i2cLoadCellFlags,
			useLoadCell0: useLoadCell0,
			useLoadCell1: useLoadCell1,
			useLoadCell2: useLoadCell2,
			useLoadCell3: useLoadCell3,
			i2cLoadCell0: i2cLoadCell0,
			i2cLoadCell1: i2cLoadCell1,
			i2cLoadCell2: i2cLoadCell2,
			i2cLoadCell3: i2cLoadCell3,
			resolvedLoadCell: resolvedLoadCell,
			loadCellZero: loadCellZero,
			loadCellSpan: loadCellSpan,
			weightSpan: weightSpan,
			weightOffset: weightOffset,
			filterMode: filterMode,
			displayWeight: displayWeight
		});
	});
	
	//__________________________________________________________________________________________
	//diag output
	
	//webpage requesting data from the pi
	socket.on('diagOutputReq', function(data){
		socket.emit('diagOutputRes',{
			i2cMainMcuFlags: i2cMainMcuFlags,
			masterOn: masterOn,
			binSensorInstalled: binSensorInstalled,
			binEmpty: binEmpty,
			swathWidth: swathWidth,
			distancePerPulse: distancePerPulse,
			pulsesPerHa: pulsesPerHa,
			productCal: productCal,
			targetRate: targetRate,
			meterSpeedRatio: meterSpeedRatio,
			i2cGroundSpeed: i2cGroundSpeed,
			gpsSpeedKph: gpsSpeedKph,
			simGroundSpeedPps: simGroundSpeedPps,
			i2cTargetMeterSpeed: i2cTargetMeterSpeed,
			i2cMeterSpeed: i2cMeterSpeed,
			i2cValve1Command: i2cValve1Command,
			manualValve1: manualValve1,
			actualRate: actualRate,
			fanInstalled: fanInstalled,
			fanRpm: fanRpm,
			simGroundSpeedKph: simGroundSpeedKph,
			gpsSpeedKph: gpsSpeedKph
		});
	});
	
	//webpage master switch pressed
	socket.on('toggleMaster', function (data){
		masterOn = !masterOn;
		
		//save data storage file
		if (masterOn == false){
			saveDataStorageFile();
		}
		
		setMeterRatio();
	});
	
	socket.on('setRate', function (data){
		targetRate = data;
		setMeterRatio();
	});
	
	socket.on('setManGroundSpd', function (data){
		//user enters simulated ground speed km/h, so calc the equivalent
		//sensor pulses per sec using the distance per pulse value
		simGroundSpeedKph = data;
		simGroundSpeedPps = simGroundSpeedKph / (distancePerPulse * 3.6);
		setMeterRatio();
	});
	
	socket.on('setManValveCommand', function (data){
		manualValve1 = data;
		setMeterRatio();
	});
	
	socket.on('stopHydTests', function (data){
		masterOn = false;
		simGroundSpeedKph = 0;
		simGroundSpeedPps = 0;
		manualValve1 = 0;
		setMeterRatio();
	});
	
	//__________________________________________________________________________________________


});

//server port
server.listen(80);






